﻿//#define SLEEP_BETWEEN_FRAME_GETS
using System;
using System.Collections;
using System.Collections.Generic;
using System.Threading;
using System.Threading.Tasks;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR.MagicLeap;

namespace Vartid.MLRecorder
{
    #region Enums
    public enum RecordingStatus
    {
        Creating,
        CreateError,
        Created,
        MissingPermission,
        Granted,
        ConnectError,
        Stopped,
        Starting,
        Error,
        Recording,
        Stopping,
    }

    [Flags]
    public enum MLPermissionFlags
    {
        None = 0,
        Camera = 1 << 0,
        RecordAudio = 1 << 1,
        ReadExternalStorage = 1 << 2,
        WriteExternalStorage = 1 << 3,
        Internet = 1 << 4,
        EyeTracking = 1 << 5,
        PupilSize = 1 << 6,
        SpatialAnchors = 1 << 7,
        HandTracking = 1 << 8,
        WebView = 1 << 9,
        MarkerTracking = 1 << 10,
        VoiceInput = 1 << 11,
        SpatialMapping = 1 << 12,
        DepthCamera = 1 << 13,
    }

    #endregion

    public class RecorderMenu : MonoBehaviour
    {
        #region CAMERA RECORDING MAIN
        private RecordingStatus recordingStatus = RecordingStatus.Creating;

        void Awake()
        {
            permissionCallbacks.OnPermissionGranted += OnPermissionGranted;
            permissionCallbacks.OnPermissionDenied += OnPermissionDenied;
            permissionCallbacks.OnPermissionDeniedAndDontAskAgain += OnPermissionDenied;
        }

        void Start()
        {
            CreateCameras();
            RequestPermissions();
        }

        private void CreateCameras()
        {
            bool depthok = CreateDepthCamera();
            //Thread.Sleep(10_000); // just wait a bit to let the timstamp differences more visible if they started from the creation time
            if (depthok &&  CreateWorldCamera())
            {
                recordingStatus = RecordingStatus.Created;
            }
            else
            {
                recordingStatus = RecordingStatus.CreateError;
            }
        }

        private void ConnectCameras()
        {
            if (ConnectDepthCamera() &&
                ConnectWorldCamera() &&
                ConnectCVCamera())
            {
                recordingStatus = RecordingStatus.Stopped;
            }
            else
            {
                recordingStatus = RecordingStatus.ConnectError;
            }
        }

        private void DisconnectCameras()
        {
            DisonnectDepthCamera();
            DisonnectWorldCamera();
            DisonnectCVCamera();
        }

        private static readonly System.Threading.Mutex startStopMutex = new();

        void StartRecording()
        {
            startStopMutex.WaitOne();
            try
            {
                if (recordingStatus == RecordingStatus.Stopped)
                {
                    recordingStatus = RecordingStatus.Recording;
                    StartVideoCapture();
                    StartDepthCapture();
                    StartWorldCapture();
                    stopButtonIsActive = true;
                }
            }
            finally
            {
                startStopMutex.ReleaseMutex();
            }
        }

        private void DoStartRecording()
        {
            Task.Run(() => StartRecording() );
        }
        private void StopRecording()
        {
            startStopMutex.WaitOne();
            try
            {
                if (recordingStatus == RecordingStatus.Recording)
                {
                    recordingStatus = RecordingStatus.Stopping;
                    StopVideoCapture();
                    StopDepthCapture();
                    StopWorldCapture();

                    recordingStatus = RecordingStatus.Stopped;
                    stopButtonIsActive = false;
                }
            }
            finally
            {
                startStopMutex.ReleaseMutex();
            }
        }

        private void DoStopRecording()
        {
            StopRecording();
        }

        #endregion

        #region PERMISSION HANDLING
        private static readonly Dictionary<string, MLPermissionFlags> requiredMLPermissions = new()
        {
            { MLPermission.Camera, MLPermissionFlags.Camera },
            { MLPermission.DepthCamera, MLPermissionFlags.DepthCamera }
        };
        private static MLPermissionFlags requiredMLPermissionFlags = MLPermissionFlags.Camera | MLPermissionFlags.DepthCamera;

        private MLPermissionFlags grantedMLPermissionFlags = MLPermissionFlags.None;
        private readonly MLPermissions.Callbacks permissionCallbacks = new();

        private void RequestPermissions()
        {
            foreach (string permission in requiredMLPermissions.Keys)
            {
                MLPermissions.RequestPermission(permission, permissionCallbacks);
            }

        }


        private void OnPermissionDenied(string permission)
        {
            if (requiredMLPermissions.ContainsKey(permission))
            {
                grantedMLPermissionFlags &= ~requiredMLPermissions[permission]; //clear Flags of MLPermission
            }
        }

        private void OnPermissionGranted(string permission)
        {
            if (requiredMLPermissions.ContainsKey(permission))
            {
                grantedMLPermissionFlags |= requiredMLPermissions[permission];   //set Flags of MLPermission
                if (recordingStatus == RecordingStatus.Created
                    && requiredMLPermissionFlags == (grantedMLPermissionFlags & requiredMLPermissionFlags)) // all permission granted
                {
                    recordingStatus = RecordingStatus.Granted;
                    ConnectCameras();
                }
            }
        }


        private void ClearPermissionCallbacks()
        {
            permissionCallbacks.OnPermissionGranted -= OnPermissionGranted;
            permissionCallbacks.OnPermissionDenied -= OnPermissionDenied;
            permissionCallbacks.OnPermissionDeniedAndDontAskAgain -= OnPermissionDenied;
        }
        #endregion

        #region DEPTH CAMERA HANDLING
        long lastDepthFrameNumber = -1;
        bool firstDepthFrame = true;
        long firstDepthSysTs;
        double lastDepthSysTs = -1d;

        MLDepthCamera.Settings depthSettings;
        private bool CreateDepthCamera()
        {
            uint longRange = (uint)MLDepthCamera.FrameType.LongRange;
            uint shortRange = (uint)MLDepthCamera.FrameType.ShortRange;
            MLDepthCamera.StreamConfig[] config = new MLDepthCamera.StreamConfig[Math.Max(longRange, shortRange) + 1];

            config[longRange].Flags = (uint)MLDepthCamera.CaptureFlags.DepthImage;
            config[longRange].Exposure = 1600; //TODO read from parameter
            config[longRange].FrameRateConfig = MLDepthCamera.FrameRate.FPS_5;

            config[shortRange].Flags = (uint)MLDepthCamera.CaptureFlags.DepthImage;
            config[shortRange].Exposure = 375; //TODO read from parameter; 400 >= max
            config[shortRange].FrameRateConfig = MLDepthCamera.FrameRate.FPS_5;

            depthSettings = new MLDepthCamera.Settings()
            {
                //Streams = MLDepthCamera.Stream.LongRange | MLDepthCamera.Stream.ShortRange,  // leads to: Failed to connect to depth camera: InvalidParam
                Streams = MLDepthCamera.Stream.LongRange,
                StreamConfig = config
            };

            MLDepthCamera.SetSettings(depthSettings);

            return true;
        }

        private bool ConnectDepthCamera()
        {
            var result = MLDepthCamera.Connect();
            if (result.IsOk && MLDepthCamera.IsConnected)
            {
                return true;
            }
            else
            {
                recordingStatus = RecordingStatus.ConnectError;
                Debug.LogError($"Failed to connect to depth camera: {result.Result}");
                return false;
            }
        }

        private void DisonnectDepthCamera()
        {
            if (MLDepthCamera.IsConnected)
            {
                MLDepthCamera.Disconnect();
            }
        }

        Task depthCaptureTask;

        CancellationTokenSource depthCaptureCTS;

        void StartDepthCapture()
        {
            depthCaptureCTS = new();
            depthCaptureTask = Task.Run(() => GetDepthFrames(depthCaptureCTS.Token), depthCaptureCTS.Token);
        }

        void GetDepthFrames(CancellationToken cts)
        {
#if SLEEP_BETWEEN_FRAME_GETS
            var stopwatch = new System.Diagnostics.Stopwatch();
            long millisecondsBetweenFrames = 1_000 / 5;
            stopwatch.Start();
#endif
            void TryGetDepthFrame()
            {
                try
                {
                    var hasnewframe = GetDepthFrame();
#if SLEEP_BETWEEN_FRAME_GETS
                    if (hasnewframe)
                    {
                        long sleepMillisecs = millisecondsBetweenFrames - stopwatch.ElapsedMilliseconds;
                        if (sleepMillisecs > 1 && sleepMillisecs <= millisecondsBetweenFrames)
                        {
                            Thread.Sleep((int)sleepMillisecs);
                        }
                        stopwatch.Restart();
                    }
                    else
                    {
                        Thread.Sleep(10);
                    }
#endif
                }
                catch (Exception e)
                {
                    Debug.LogError($"Error in GetDepthFrames: {e.Message}");
                }
            }

            while (!cts.IsCancellationRequested)
            {
                TryGetDepthFrame();
            }
#if SLEEP_BETWEEN_FRAME_GETS
            stopwatch.Stop();
#endif
        }

        void StopDepthCapture()
        {
            if (depthCaptureCTS != null)
            {
                depthCaptureCTS.Cancel();
            }
        }

        bool GetDepthFrame()
        {
            bool frameSaved = false;
            try
            {
                if (recordingStatus == RecordingStatus.Recording)
                {
                    if (MLDepthCamera.IsConnected)
                    {
                        MLResult result = MLDepthCamera.GetLatestDepthData(0, out MLDepthCamera.Data data);
                        if (result.IsOk && data != null && data.FrameNumber != lastDepthFrameNumber)
                        {
                            lastDepthFrameNumber = data.FrameNumber;
                            MLTime.ConvertMLTimeToSystemTime(data.FrameTimestamp, out long sysTimeStamp);
                            lastDepthSysTs = sysTimeStamp / 1_000_000_000d;
                            //infoStr = $"Depth@{lastDepthSysTs:F3}, CV@{lastCvSysTs:F3}, World/2@{lastWorldSysTs/2:F3}, World@{lastWorldSysTs:F3}";
                            //Debug.Log(infoStr);
                            frameSaved = true;
                        }
                    }
                }

            }
            catch (Exception e)
            {
                Debug.LogError($"Error on GetDepthframe: {e.Message}");
            }
            return frameSaved;
        }

#endregion

#region CV CAMERA HANDLING
        private int cvWidth = 1280;
        private int cvHeight = 960;
        private long lastCVFrameNumber = -1;
        double lastCvSysTs = -1d;

        MLCamera.StreamCapability cvStreamCapability;

        private MLCamera cvCamera = null;
        private static readonly MLCamera.CaptureType cvCaptureType = MLCamera.CaptureType.Video; // Do not change!
        private static readonly MLCamera.OutputFormat cvOutputFormat = MLCamera.OutputFormat.RGBA_8888; // Do not change!

        private bool ConnectCVCamera()
        {
            StartCoroutine(EnableMLCamera());

            MLCamera.ConnectContext context = MLCamera.ConnectContext.Create();
            context.Flags = MLCameraBase.ConnectFlag.CamOnly;
            context.EnableVideoStabilization = false;

            cvCamera = MLCamera.CreateAndConnect(context);

            if (cvCamera != null)
            {
                if (CheckCVCameraStreamCapabilities())
                {
                    cvCamera.OnRawVideoFrameAvailable += OnCaptureRawVideoFrameAvailable;
                    return true;
                }
                else
                {
                    Debug.LogError("CV Camera device DID NOT receive stream capabilities.");
                }
            }
            Debug.LogError("Failed to connect to CV camera.");
            recordingStatus = RecordingStatus.ConnectError;
            return false;
        }

        private bool cvCameraIsAvailable = false;

        private IEnumerator EnableMLCamera()
        {
            while (!cvCameraIsAvailable)
            {
                MLResult result =
                    MLCamera.GetDeviceAvailabilityStatus(MLCamera.Identifier.Main, out cvCameraIsAvailable);
                if (!(result.IsOk && cvCameraIsAvailable))
                {
                    // Wait until camera device is available
                    yield return new WaitForSeconds(1.0f);
                }
            }
        }

        private void DisonnectCVCamera()
        {
            if (cvCamera != null && cvCamera.ConnectionEstablished)
            {
                cvCamera.OnRawVideoFrameAvailable -= OnCaptureRawVideoFrameAvailable;
                cvCamera.Disconnect();
            } else
            {
                // Note that some APIs like MLCameraInit() can be called before MLCameraConnect()
                // is called. This is to make sure all is cleaned up if CameraConnect is not called
                MLCamera.Uninitialize();
            }
        }

        private bool CheckCVCameraStreamCapabilities()
        {
            var result = cvCamera.GetStreamCapabilities(out MLCamera.StreamCapabilitiesInfo[] streamCapabilitiesInfo);
            if (!result.IsOk)
            {
                Debug.LogError("Could not get Stream capabilities Info.");
                return false;
            }

            for (int i = 0; i < streamCapabilitiesInfo.Length; i++)
            {
                foreach (var streamCap in streamCapabilitiesInfo[i].StreamCapabilities)
                {
                    if (streamCap.Height == cvHeight && streamCap.Width == cvWidth)
                    {
                        cvStreamCapability = streamCap;
                        return true;
                    }
                }
            }
            return false;
        }

        MLCamera.Metadata cvMetaData;
        private void StartVideoCapture()
        {
            MLCamera.CaptureConfig captureConfig = new();
            captureConfig.CaptureFrameRate = MLCamera.CaptureFrameRate._15FPS;
            captureConfig.StreamConfigs = new MLCamera.CaptureStreamConfig[1];
            captureConfig.StreamConfigs[0] = MLCamera.CaptureStreamConfig.Create(cvStreamCapability, cvOutputFormat);

            MLResult result = cvCamera.PrepareCapture(captureConfig, out cvMetaData);
            if (MLResult.DidNativeCallSucceed(result.Result, nameof(cvCamera.PrepareCapture)))
            {
                cvCamera.PreCaptureAEAWB();
                int[][] regions = new int[1][] {
                     new int[5] { 3 * cvWidth / 8, 3 * cvHeight / 8, 5 * cvWidth / 8, 5 * cvHeight / 8, 10000}, // 1/16 center AERegion::kMaxPossibleWeight
                    };
                cvMetaData.SetControlAERegions(regions);

                cvMetaData.SetControlAEExposureCompensation(0);

                cvCamera.UpdateCaptureSettings();

                result = cvCamera.CaptureVideoStart();
                if (!MLResult.DidNativeCallSucceed(result.Result, nameof(cvCamera.CaptureVideoStart)))
                {
                    recordingStatus = RecordingStatus.Error;
                }
            }
        }

        private void StopVideoCapture()
        {
            cvCamera.CaptureVideoStop();
        }


        private void OnCaptureRawVideoFrameAvailable(MLCamera.CameraOutput capturedFrame, MLCamera.ResultExtras resultExtras, MLCamera.Metadata metadataHandle)
        {
            if (recordingStatus == RecordingStatus.Recording && resultExtras.FrameNumber != lastCVFrameNumber)
            {
                try
                {
                    Matrix4x4 cameraTransform = new();
                    if (MLCVCamera.GetFramePose(resultExtras.VCamTimestamp, out cameraTransform).IsOk)
                    {
                        lastCVFrameNumber = resultExtras.FrameNumber;
                        MLTime.ConvertMLTimeToSystemTime(resultExtras.VCamTimestamp.Value, out long sysTimeStamp);
                        lastCvSysTs = sysTimeStamp  / 1_000_000_000d;
                        //infoStr = $"Depth@{lastDepthSysTs:F3}, CV@{lastCvSysTs:F3}, World/2@{lastWorldSysTs / 2:F3}, World@{lastWorldSysTs:F3}";
                        //Debug.Log(infoStr);
                    }
                }
                catch (Exception e)
                {
                    Debug.LogError($"CV frame Dropped in OnCaptureRawVideoFrameAvailable: {e.Message}");
                }
            }
        }
#endregion

#region WORLD CAMERA HANDLING
        private long lastWorldFrameNumber = -1;
        double lastWorldSysTs = -1d;

        private MLWorldCamera worldCamera = null;

        private bool CreateWorldCamera()
        {
            worldCamera = new MLWorldCamera();
            return worldCamera != null;
        }

        private bool ConnectWorldCamera()
        {
            var settings = new MLWorldCamera.Settings(MLWorldCamera.Mode.NormalExposure, MLWorldCamera.CameraId.Center);

            var result = worldCamera.Connect(in settings);
            var retVal = result.IsOk && worldCamera.IsConnected;
            if (!retVal)
            {
                recordingStatus = RecordingStatus.ConnectError;
                Debug.LogError($"Failed to connect to World camera: {result.Result}");
            }
            return retVal;
        }

        private void DisonnectWorldCamera()
        {
            if (worldCamera != null && worldCamera.IsConnected)
            {
                worldCamera.Disconnect();
            }
        }

        Task worldCaptureTask;

        CancellationTokenSource worldCaptureCTS;

        void StartWorldCapture()
        {
            worldCaptureCTS = new();
            worldCaptureTask = Task.Run(() => GetWorldFrames(worldCaptureCTS.Token), worldCaptureCTS.Token);
        }

        void GetWorldFrames(CancellationToken cts)
        {
#if SLEEP_BETWEEN_FRAME_GETS
            long millisecondsBetweenFrames = 1_000 / 10; /*1000 / worldPerSec*/
            var stopwatch = new System.Diagnostics.Stopwatch();
            stopwatch.Start();
#endif
            while (!cts.IsCancellationRequested)
            {
                try
                {
                    var hasnewframe = GetWorldFrame();
#if SLEEP_BETWEEN_FRAME_GETS
                    if (hasnewframe)
                    {
                        long sleepMillisecs = millisecondsBetweenFrames - stopwatch.ElapsedMilliseconds;
                        Thread.Sleep((int)Math.Min(Math.Max(1, sleepMillisecs), millisecondsBetweenFrames));
                        stopwatch.Restart();
                    }
                    else
                    {
                        Thread.Sleep(10);
                    }
#endif
                }
                catch (Exception e)
                {
                    Debug.LogError($"in GetDepthFrames: {e.Message}");
                }
            }
#if SLEEP_BETWEEN_FRAME_GETS
            stopwatch.Stop();
#endif
        }

        void StopWorldCapture()
        {
            if (worldCaptureCTS != null)
            {
                worldCaptureCTS.Cancel();
            }
        }

        bool GetWorldFrame()
        {
            bool frameSaved = false;
            try
            {
                if (recordingStatus == RecordingStatus.Recording)
                {
                    if ( worldCamera != null && worldCamera.IsConnected )
                    {
                        MLResult result = worldCamera.GetLatestWorldCameraData(out MLWorldCamera.Frame[] frames);
                        if (result.IsOk && frames != null)
                        {
                            foreach (MLWorldCamera.Frame frame in frames)
                            {
                                if (lastWorldFrameNumber != frame.FrameNumber)
                                {
                                    lastWorldFrameNumber = frame.FrameNumber;
                                    MLTime.ConvertMLTimeToSystemTime(frame.TimeStamp, out long sysTimeStamp);
                                    lastWorldSysTs = sysTimeStamp / 1_000_000_000d;
                                    //infoStr = $"Depth@{lastDepthSysTs:F3}, CV@{lastCvSysTs:F3}, World/2@{lastWorldSysTs / 2:F3}, World@{lastWorldSysTs:F3}";
                                    //Debug.Log(infoStr);
                                    frameSaved = true;
                                }
                            }
                        }
                    }
                }
            }
            catch (Exception e)
            {
                Debug.LogError($"on GetWorldFrame: {e.Message}");
            }
            return frameSaved;
        }

#endregion

#region UI
        public GameObject stopButton;
        public Text statusText;
        //private string infoStr = "";

        private bool stopButtonIsActive = false;


        public void OnPressStartRecordingButton()
        {
            DoStartRecording();
        }

        public void OnPressStoptRecordingButton()
        {
            DoStopRecording();
        }
        public void OnPressQuitButton()
        {
            Application.Quit();
        }

        double c_d = 0;
        double c_d_min = double.MaxValue;
        double c_d_max = double.MinValue;
        double w_d = 0;
        double w_d_min = double.MaxValue;
        double w_d_max = double.MinValue;
        string c_info = "";
        string w_info = "";
        string info = " ";
        void Update()
        {
            if (lastDepthSysTs > 0)
            {
                if (lastCvSysTs >0)
                {
                    c_d = lastCvSysTs- lastDepthSysTs;
                    c_d_min = Math.Min(c_d_min, c_d);
                    c_d_max = Math.Max(c_d_max, c_d);
                    c_info = $"Color - Depth: {c_d:F3}, Min: {c_d_min:F3}, Max: {c_d_max:F3}\n";
                }
                if (lastWorldSysTs > 0)
                {
                    w_d = lastWorldSysTs/2d- lastDepthSysTs;
                    w_d_min = Math.Min(w_d_min, w_d);
                    w_d_max = Math.Max(w_d_max, w_d);
                    w_info = $"Wolrd/2 - Depth: {w_d:F3}, Min: {w_d_min:F3}, Max: {w_d_max:F3}";
                }

                info = $"Depth SysTs:@{lastDepthSysTs:F3}\n" +
                $"Color SysTs@{lastCvSysTs:F3}\n" +
                $"World SysTs/2@{lastWorldSysTs / 2:F3}\n" +
                $"World SysTs@{lastWorldSysTs:F3}\n" +
                c_info + w_info;
            }
            statusText.text = info;

            stopButton.SetActive(stopButtonIsActive);
        }

#endregion

        private void OnDestroy()
        {
            ClearPermissionCallbacks();
            DisconnectCameras();
        }

    }

}